<?php

namespace App\Services;

use App\Helpers\Helpers;
use mysql_xdevapi\Exception;
use WHMCS\Database\Capsule;
use App\Helpers\ConfigReader;
use App\Helpers\Logger;

// TODO put this somewhere better
function getCurrency($currencyCode, $systemCurrencies) {

    $thisCurrency = null;

    foreach ($systemCurrencies as $sCurrency) {
        if ($sCurrency['code'] == $currencyCode) {
            $thisCurrency = $sCurrency;
        }
    }

    return $thisCurrency;
}

class WHMCSRepository
{

    const SETTINGS = [        
        'api_user',
        'api_password',
        'api_base_url',
        'log_type',
        'log_stream',
        'log_path',
        'test_mode',
        'log_api_requests',
        'subordinate_ca',
        'dv_identity',
        'ev_identity',
        'ov_identity',
        'codesign_identity',
        "pac_identity",
        'google_maps_api_key',
        'show_sectigo_products_page'
    ];

    const FORM_FIELD_VALUES_NAME = 'form_field_values_encoded';

    public static function getApiCredentials()
    {
        if (!Capsule::schema()->hasTable('mod_sectigo_settings')) {
            return [];
        }

        $settings = [];
        $db_settings = Capsule::table('mod_sectigo_settings')->get();
        foreach (WHMCSRepository::SETTINGS as $setting_name) {
            foreach ($db_settings as $db_setting) {
                if ($db_setting->setting == $setting_name) {
                    $settings[$setting_name] = $db_setting->value;
                }
            }
        }
        $settings['api_password'] = decrypt($settings['api_password']);
        $settings['test_mode'] = $settings['test_mode']  ? 'checked' : '';
        $settings['log_api_requests'] = $settings['log_api_requests']  ? 'checked' : '';

        return $settings;
    }

    public static function insertApiCredentials()
    {
        if (!Capsule::schema()->hasTable('mod_sectigo_settings')) {
            throw new \Exception("Database error! No sectigo products table!", 500);
        }

        foreach (WHMCSRepository::SETTINGS as $setting_name) {
            $value = $_POST[$setting_name];
            if ($setting_name == 'api_password') {
                $value = encrypt($value);
            } elseif ($setting_name == 'test_mode') {
                $value = isset($value) ? 1 : 0;
            } elseif ($setting_name == 'log_api_requests' || $setting_name == 'subordinate_ca' || $setting_name == 'show_sectigo_products_page') {
                $value = isset($value) ? 1 : 0;
            }
            WHMCSRepository::setConfigSetting($setting_name, $value);

        }
    }

    public static function getConfigSetting($setting)
    {
        $db_setting = Capsule::table('mod_sectigo_settings')
            ->where('setting', $setting)
            ->first();
        if ($db_setting) {
            return $db_setting->value;
        } else {
            return null;
        }
    }

    /*
     * Sets the version of the addon module installed in the settings table
     */
    public static function setConfigSetting($setting, $value) 
    {
        $db_setting = Capsule::table('mod_sectigo_settings')
            ->where('setting', $setting)
            ->first();
        if ($db_setting) {
            Capsule::table('mod_sectigo_settings')
                ->where('setting', $setting)
                ->update([
                    'value' => $value
                ]);
        } else {
            Capsule::table('mod_sectigo_settings')->insert([
                'setting' => $setting,
                'value' => $value,
            ]);
        }
    } 

    /**
     * @param int $orderid
     * @param int $omsnumber
     * @param int $serviceid
     * @throws \Exception
     * @deprecated - use upsertServiceDetail instead.
     */
    public static function insertServiceDetail($orderid, $omsnumber, $serviceid) {

        Capsule::table('mod_sectigo_service_details')->insert(
            [
                'orderid'    => $orderid,
                'omsnumber'  => $omsnumber,
                'serviceid'  => $serviceid,
                'created_at' => new \DateTime(),
                'updated_at' => new \DateTime(),
            ]
        );
    }

    public static function upsertModServiceDetail($orderid, $omsnumber, $serviceid, $status='Pending') {

        $now = date("Y-m-d H:i:s");

        $id = Capsule::table('mod_sectigo_service_details')
                ->where('serviceid', $serviceid)
                ->where('orderid', $orderid)
                ->value('id');

        if ($id) {

            $updateRow = [
                'omsnumber'  => $omsnumber,
                'updated_at' => $now,
                'status'     => $status,
            ];

            Capsule::table('mod_sectigo_service_details')
                ->where('id', $id)
                ->update($updateRow);

        } else {

            $insertRow = [
                'created_at' => $now,
                'orderid'    => $orderid,
                'omsnumber'  => $omsnumber,
                'serviceid'  => $serviceid,
                'status'     => $status
            ];

            Capsule::table('mod_sectigo_service_details')->insert($insertRow);

        }

    }

    public static function upsertServiceDetail($orderid, $omsnumber, $serviceid, $status='Pending') {
        $pdo = Capsule::connection()->getPdo();
        $pdo->beginTransaction();

        try {
            $statement = $pdo->prepare(
                'insert into mod_sectigo_service_details (orderid, omsnumber, status, serviceid, created_at) values (:orderid, :omsnumber, :status, :serviceid, :dateCreated) ON DUPLICATE KEY UPDATE omsnumber = :newomsnumber, updated_at= :dateUpdated, status = :newstatus'
            );

            $statement->execute(
                [
                    ':orderid'      => $orderid,
                    ':omsnumber'    => $omsnumber,
                    ':status'       => $status,
                    ':serviceid'    => $serviceid,
                    ':dateCreated'  => date("Y-m-d H:i:s"),
                    ':newomsnumber' => $omsnumber,
                    ':dateUpdated'  => date("Y-m-d H:i:s"),
                    ':newstatus'       => $status,
                ]
            );

            $pdo->commit();
        } catch (\Exception $e) {
            $pdo->rollBack();
            throw new \Exception('Database error: ' .$e->getMessage());
        }
    }

    public static function upsertAddonDetail($orderid, $omsnumber, $addonid, $status='Pending') {
        $pdo = Capsule::connection()->getPdo();
        $pdo->beginTransaction();

        try {
            $statement = $pdo->prepare(
                'insert into mod_sectigo_addon_details (orderid, omsnumber, status, hostingaddonid, created_at)
                 values (:orderid, :omsnumber, :status, :hostingaddonid, :dateCreated) ON DUPLICATE KEY UPDATE
                 omsnumber = :newomsnumber, updated_at= :dateUpdated, status = :newstatus'
            );

            $statement->execute(
                [
                    ':orderid'      => $orderid,
                    ':omsnumber'    => $omsnumber,
                    ':status'       => $status,
                    ':hostingaddonid'    => $addonid,
                    ':dateCreated'  => date("Y-m-d H:i:s"),
                    ':newomsnumber' => $omsnumber,
                    ':dateUpdated'  => date("Y-m-d H:i:s"),
                    ':newstatus'       => $status,
                ]
            );

            $pdo->commit();
        } catch (\Exception $e) {
            $pdo->rollBack();
            throw new \Exception('Database error: ' .$e->getMessage());
        }
    }

    public static function updateHostingAddon($hostingaddonid, $props)
    {
        Capsule::table('tblhostingaddons')
            ->where('id', '=', $hostingaddonid)
            ->update($props);
    }

    public static function setServiceDetails($serviceID, $externalOrderId, string $status)
    {
        $entry = Capsule::table('mod_sectigo_service_details')->where('serviceid', $serviceID);
        $entry->update([
            'status'     => $status,
            'omsnumber'    => $externalOrderId,
            'updated_at' => new \DateTime()
        ]);
        // Make sure custom field status stays in sync
        self::setProductServiceCustomFields($serviceID, ['status' => $status]);
    }

    public static function setAddonDetails($addonID, $externalOrderId, string $status)
    {
        $entry = Capsule::table('mod_sectigo_addon_details')->where('hostingaddonid', $addonID);
        $entry->update([
            'status'     => $status,
            'omsnumber'    => $externalOrderId,
            'updated_at' => new \DateTime()
        ]);
    }

	/**
	 * Sets the service's status in the database. Updates both mod_sectigo_service_details table and the custom fields.
	 *
	 * @param int    $serviceID The service's ID.
	 * @param string $status    The status to set.
	 *
	 * @return void
	 */
    public static function setServiceStatus(int $serviceID, string $status) : void
    {
        $entry = Capsule::table('mod_sectigo_service_details')->where('serviceid', $serviceID);
        $entry->update([
            'status'     => $status,
            'updated_at' => new \DateTime()
        ]);
        // Make sure custom field status stays in sync
        self::setProductServiceCustomFields($serviceID, ['status' => $status]);
    }

    public static function setAddonStatus($addonID, string $status)
    {
        $entry = Capsule::table('mod_sectigo_addon_details')->where('hostingaddonid', $addonID);
        $entry->update([
            'status'     => $status,
            'updated_at' => new \DateTime()
        ]);
    }

    public static function getModServiceDetail($serviceID) {
        return Capsule::table('mod_sectigo_service_details')->where('serviceid', $serviceID)->first();
    }

    public static function getModAddonDetail($addonID) {
        return Capsule::table('mod_sectigo_addon_details')->where('hostingaddonid', $addonID)->first();
    }

    public static function getServiceDetail($serviceID) {
        return Capsule::table('tblhosting')->where('id', $serviceID)->first();
    }

    public static function getAddonDetail($addonId) {
        return Capsule::table('tblhostingaddons')->where('id', $addonId)->first();
    }

    public static function getServiceStatus($serviceID) {
        $mod_service = self::getModServiceDetail($serviceID);
        $order       = self::getOrderDetail($mod_service->orderid);
        $service     = self::getServiceDetail($serviceID);

        if ($mod_service && $mod_service->id) {
            return [
                'orderId' => $mod_service->orderid,
                'orderStatus' => $order->status,
                'serviceId' => $mod_service->serviceid,
                'serviceStatus' => $service->domainstatus,
                'provisioningStatus' => $mod_service->status,
                'omsNumber' => $mod_service->omsnumber,
            ];
        } else {
            return [];
        }
    }

    public static function getAddonStatus($addonId) {
        $mod_addon   = self::getModAddonDetail($addonId);
        $order       = self::getOrderDetail($mod_addon->orderid);
        $addon       = self::getAddonDetail($addonId);
        $service     = self::getServiceDetail($addon->hostingid);

        if ($mod_addon && $mod_addon->id) {
            return [
                'orderId' => $mod_addon->orderid,
                'orderStatus' => $order->status,
                'addonId' => $mod_addon->hostingaddonid,
                'addonStatus' => $addon->status,
                'serviceStatus' => $service->status,
                'provisioningStatus' => $mod_addon->status,
                'omsNumber' => $service->omsnumber,
            ];
        } else {
            return [];
        }
    }

    public static function getAddonsForService($serviceId, $status) {
        return Capsule::table('mod_sectigo_addon_details')
            ->join('tblhostingaddons', 'tblhostingaddons.id', '=', 
                'mod_sectigo_addon_details.hostingaddonid')
            ->where('tblhostingaddons.hostingid', $serviceId)
            ->where('mod_sectigo_addon_details.status', $status)
            ->get();
    }

    /*
     * Creates or sets the tblcustomfieldsvalues entries for the
     * fieldValues (key=>val) array.  If the custom field does not exist,
     * it creates one.
     */
    public static function setProductServiceCustomFields($serviceID, $fieldValues, $type='product')
    {
        $pid = Capsule::table('tblhosting')
            ->where('id', $serviceID)
            ->value('packageid');
        $idValues = [];
        foreach ($fieldValues as $field => $value) {
            $customFieldId = Capsule::table('tblcustomfields')
                ->where('relid', $pid)
                ->where('type', $type)
                ->whereRaw('(fieldname LIKE ? or fieldname = ?)',
                    [$field.'|%', $field])
                ->value('id');
            if ($customFieldId) {
                $idValues[$customFieldId] = $value;
            } else {
                $customFieldId = Capsule::table('tblcustomfields')->insertGetId([
                        'type'=> $type,
                        'relid' => $pid,
                        'fieldname' => $field,
                        'fieldtype'=>'text',
                        'adminonly' => 'on',
                        'sortorder' => 0,
                    ]);
                $idValues[$customFieldId] = $value;
            }
        }
        foreach ($idValues as $id => $value) {
            $fieldValueId = Capsule::table('tblcustomfieldsvalues')
                ->where('fieldid', $id)
                ->where('relid', $serviceID)
                ->value('id');
            if ($fieldValueId) {
                Capsule::table('tblcustomfieldsvalues')
                    ->where('id', $fieldValueId)
                    ->update(['value' => $value, 'updated_at' => date('c')]);
            } else {
                Capsule::table('tblcustomfieldsvalues')->insert([
                    'fieldid'=>$id,
                    'relid' => $serviceID,
                    'value' => $value,
                    'created_at' => date('c'),
                    'updated_at' => date('c'),
                ]);
            }
        }
    }

    /*
     * v1.2.3 Deprecated functionality.  Use setProductServiceCustomFields
     */
    public static function setProductServiceCustomFields2($fieldValues, $serviceId, $productId) {

        $customFields = Capsule::table('tblcustomfields')
                        ->where('relid', $productId)
                        ->where('type', 'product')->get();
        $ids = [];
        $errors = [];

        foreach ($fieldValues as $k => $v) {

            $matchedField = null;
            foreach($customFields as $cf) {
                if (strstr(strtolower(explode("|", $cf->fieldname)[0]), strtolower($k))) {
                    $matchedField = $cf;
                    break;
                }
            }

            if ($matchedField->id) {

                $customFieldValue = ['value' => $v,  
                                     'updated_at' => date('c'),
                                     'relid' => $serviceId, 
                                     'fieldid' => $matchedField->id,];

                try {
                    $id = Capsule::table('tblcustomfieldsvalues')->insertGetId($customFieldValue);
                    $ids[] = $id;
                } catch (\Exception $e) {
                    $errors[] = "cannot insert to tblcustomfieldsvalues with values " . print_r($customFieldValue, true);
                }
            }
        }

        if (count($errors) == 0) {
            return ['status' => 'success',
                    'data'   => $ids
            ];
        } else {
            return ['status' => 'error',
                    'errors' => $errors,
            ];
        }
    }

    /*
     * v1.2.3 Deprecated functionality.  Use setProductServiceCustomFields
     */
    public static function updateProductServiceCustomFields($fieldValues, $orderId, $productId) {
        $customFields = Capsule::table('tblcustomfields')
                        ->where('relid', $productId)
                        ->where('type', 'product')->get();
        $ids = [];
        $errors = [];

        foreach ($fieldValues as $k => $v) {

            $matchedField = null;
            foreach($customFields as $cf) {
                if (strstr(strtolower(explode("|", $cf->fieldname)[0]), strtolower($k))) {
                    $matchedField = $cf;
                    break;
                }
            }

            if ($matchedField->id) {

                $customFieldValue = ['value' => $v,  
                                     'updated_at' => date('c')];
                $id = Capsule::table('tblcustomfieldsvalues')
                        ->where('relid', $orderId)
                        ->where('fieldid', $matchedField->id)
                        ->update($customFieldValue);
                $ids[] = $id;
                
            }
        }
    }

    public static function getOrderDetail($orderID) {
        return Capsule::table('tblorders')->where('id', $orderID)->first();
    }

    public static function getSSLServiceStatus($serviceId) {
        if ($serviceId) {
            return  WHMCSRepository::getService($serviceId);
        } else {
            return null;
        }
    }

    public static function getService($serviceId) {
        return Capsule::table('tblhosting')
                 ->where('id', $serviceId)
                 ->first();
    }

    public static function getAddon($addonId) {
        $addon = Capsule::table('tbladdons')->where('id', $addonId)->first();
        if ($addon) {
            $module_configs = Capsule::table('tblmodule_configuration')
                ->where('entity_id', $addonId)
                ->where('entity_type', 'addon')->get();
            $addon->configs = [];
            foreach ($module_configs as $module_config) {
                $addon->configs[$module_config->setting_name] = $module_config->value;
            }
        }
        return $addon;
    }

    public static function getAddonPricing($addonId, $currencyId, $billingcycle) {
        return Capsule::table('tblpricing')
            ->where('type', 'addon')
            ->where('relid', $addonId)
            ->where('currency', $currencyId)
            ->value(strtolower($billingcycle));
    }

    public static function deleteAddonModuleTables()
    {
        try {
            Capsule::schema()->drop('mod_sectigo_products');
            Capsule::schema()->drop('mod_sectigo_settings');
            Capsule::schema()->drop('mod_sectigo_service_details');
        } catch (\Exception $e) {
            echo "Unable to delete mod_sectigo tables: {$e->getMessage()}";
        }
        //TODO: Check does DB needs more cleaning when deactivating addon module
    }

    public static function getProductsFromCustomTable()
    {
        $products = Capsule::table('mod_sectigo_products')->get();
        return $products;
    }

    public static function initializeProductsGroup() {
        $sectigoProductGroup = Capsule::table('tblproductgroups')
            ->where('name', 'Sectigo Products')
            ->get();

        if (!$sectigoProductGroup) {
            $groupId = Capsule::table('tblproductgroups')->insertGetId([
                'name'             => 'Sectigo Products',
                'headline'         => '',
                'tagline'          => '',
                'orderfrmtpl'      => '',
                'disabledgateways' => '',
                'hidden'           => 0,
                'order'            => 1,
            ]);

            WHMCSRepository::setConfigSetting('sectigo_products_group_id', $groupId);
        }
    }

    /**
     * @param array $productsFromConfigFile
     * @throws \Exception
     */
    public static function PopulateCustomProductsTable(array $productsFromConfigFile)
    {
        if (Capsule::schema()->hasTable('mod_sectigo_products')) {

            foreach ($productsFromConfigFile as $product) {
                Capsule::table('mod_sectigo_products')->insert(
                    [
                        'name'                => $product['name'],
                        'provisioning_module' => $product['config']['provisioning_module']['name']
                    ]
                );
            }
        } else {
            throw new \Exception("Database error! No sectigo products table!", 500);
        }
    }

    /**
     * @param array $productParams
     * @param int $productGroupId
     * @throws \Exception
     *
     * https://developers.whmcs.com/api-reference/addproduct/
     */
    public static function createProduct($productParams, $productGroupId)
    {
        $command = 'AddProduct';
        $postData = array(
            'type'          => 'other',
            'gid'           => $productGroupId,
            'name'          => $productParams['name'],
            'description'   => $productParams['name'],
            'module'        => $productParams['config']['provisioning_module']['name'],
            'autosetup'     => 'payment', //When should the product be automatically setup. One of “ (never), ‘on’ (pending order), ‘payment’ (on payment), ‘order’ (on order)
            'paytype'       => $productParams['config']['pricing']['type'], //‘free’, ‘onetime’, ‘recurring’
            'showdomainoptions' => false,
        );

        /**
         * Sets products settings (Id, MultiDomain, WildCard, etc.) not Configurable Options
         */
        $optionStep = 1;
        foreach ($productParams['config']['provisioning_module']['settings'] as $setting){
            $configOptionName = 'configoption'. $optionStep;
            $postData[$configOptionName]= $setting;
            $optionStep++;
        }

        $response = localAPI($command, $postData);
        if ($response === 'success') {
            throw new \Exception('Unable to create product: ' . $productParams['name'] . '.');
        }

        if ($productParams['config']['addons']['form_names']) {
            foreach ($productParams['config']['addons']['form_names'] as $addon) {
                $addonId = self::createAddonIfNotExist($addon);
                self::addAddonToProduct($addonId, $response['pid']);
            }
        }
    }

    /**
     * Create the addon from the configuration in the database if it doesn't exist yet.
     *
     * Returns the id of the addon
     */
    public static function createAddonIfNotExist($addonName)
    {
        $configReader = new ConfigReader();
        $addonConfig = $configReader->getAddonByFormName($addonName);

        $existingAddon = Capsule::table('tbladdons')
            ->where('name', $addonConfig['name'])->first();
        if ($existingAddon) {
            return $existingAddon->id;
        }
        $currency = Capsule::table('tblcurrencies')->orderBy('default', 'desc')
            ->first();

        $addon = [
            'name' => $addonConfig['name'],
            'description' => $addonConfig['description'],
            'autoactivate' => 'payment',
            'tax' => 1,
            'showorder' => 1,
            'type' => 'other',
        ];
        $pricing = [
            'type' => 'addon',
            'currency' => $currency->id,
        ];
        if ($addonConfig['config']) {
            if ($addonConfig['config']['pricing']) {
                $addon['billingcycle'] = $addonConfig['config']['pricing']['type'];
                $pricing = array_merge($pricing, $addonConfig['config']['pricing']['values']);
            }
            if ($addonConfig['config']['provisioning_module']) {
                $provMod = $addonConfig['config']['provisioning_module'];
                $addon['module'] = $provMod['name'];
                $addon['weight'] = $provMod['weight'];
            }
        }
        $addonId = Capsule::table('tbladdons')->insertGetId($addon);
        
        $pricing['relid'] = $addonId;
        Capsule::table('tblpricing')->insert($pricing);

        if ($addonConfig['config']['provisioning_module']) {
            $provMod = $addonConfig['config']['provisioning_module'];
            foreach($provMod['settings'] as $setting) {
                Capsule::table('tblmodule_configuration')->insert(
                    array_merge($setting, [
                        'entity_type' => 'addon',
                        'entity_id' => $addonId
                    ])
                );
            }
        }

        return $addonId;
    }

    /**
     * Associates the addon with the product
     */
    public static function addAddonToProduct($addonId, $productId)
    {
        $addon = Capsule::table('tbladdons')->where('id', $addonId)->first();
        $packages = explode(',', $addon->packages);
        
        if (strlen(trim($packages[0])) == 0) {
            $packages = [$productId];
        } else if (!in_array(strval($productId), $packages)) {
            array_push($packages, strval($productId));
        }
        Capsule::table('tbladdons')->where('id', $addonId)->update([
            'packages' => implode(',', $packages)
        ]);
    }

    public static function getSectigoProductGroupId()
    {
        $sectigoProductGroupId = Capsule::table('mod_sectigo_settings')->where('setting', 'sectigo_products_group_id')->first()->value;
        $group = Capsule::table('tblproductgroups')->where('id', $sectigoProductGroupId)->get()[0];
        if($group){
            return $group->id;
        } else {
            $groupId = Capsule::table('tblproductgroups')->insertGetId([
                'name'             => 'Sectigo Products',
                'headline'         => '',
                'tagline'          => '',
                'orderfrmtpl'      => '',
                'disabledgateways' => '',
                'hidden'           => 0,
                'order'            => 1,
            ]);

            WHMCSRepository::setConfigSetting('sectigo_products_group_id', $groupId);
            return $groupId;
        }
    }

    public static function markProductAsCreated($productName)
    {
        $entry = Capsule::table('mod_sectigo_products')->where('name', $productName);
        $entry->update(['created' => true]);
    }

    public static function markProductAsCreated2($pid, $productName, $productModule)
    {
        $entry = Capsule::table('mod_sectigo_products')->where('name', '=', $productName)->first();

        if ($entry) {
            Capsule::table('mod_sectigo_products')->where('name', '=', $productName)->update(['created' => true, 'product_id' => $pid]);
        } else {
            Capsule::table('mod_sectigo_products')
                ->insert(['name' => $productName,
                          'created' => true,
                          'provisioning_module' => $productModule,
                          'product_id' => $pid]);
        }

    }

    public static function deleteSectigoProducts()
    {
        $pgid = WHMCSRepository::getSectigoProductGroupId();
        Capsule::table('tblproducts')->where('gid', $pgid)->delete();
        Capsule::table('tblproductgroups')->where('id', $pgid)->delete();
    }


    public static function getSectigoProducts($groupId)
    {
        $products = Capsule::table('tblproducts')->where('gid', $groupId)->get();

        return $products;
    }

    public static function getProduct($packageId)
    {
        return Capsule::table('tblproducts')->where('id', $packageId)->first();
    }

    public static function checkAreProductsProvisioned()
    {
        $productGroupID = WHMCSRepository::getSectigoProductGroupId();
        $products = WHMCSRepository::getSectigoProducts($productGroupID);
        foreach ($products as $product) {
            $purchased = Capsule::table('tblhosting')->where('packageid', $product->id)->get();
            if (!empty($purchased)) {
                return true;
            }
        }
        return false;
    }

    /*
     * Gets the configuration option specified for the service IDs
     *
     * @param $serviceIds   => Service IDs to get options for
     * @param $configOption => Name of the configuration option to get
     *
     * Returns array [$serviceId => 'configOptionValue']
     */
    public static function getServicesConfigOptions($serviceIds, $configOption) {
        $customFieldValues = Capsule::table('tblcustomfieldsvalues')
            ->select([
                'tblcustomfieldsvalues.relid',
                'tblcustomfieldsvalues.value'
            ])
            ->join('tblcustomfields', 'tblcustomfieldsvalues.fieldid',
                '=', 'tblcustomfields.id')
            ->where('tblcustomfields.type','=','product')
            ->whereIn('tblcustomfieldsvalues.relid', $serviceIds)
            ->whereRaw('(tblcustomfields.fieldname LIKE ? or tblcustomfields.fieldname = ?)',
                [$configOption.'|%', $configOption])
            ->get();
        $results = [];
        foreach ($customFieldValues as $customFieldValue) {
            $results[$customFieldValue->relid] = $customFieldValue->value;
        }
        return $results;
    }

    /**
     * @param int $productID
     * @param array $productFromConfigFile
     *
     * Adds Custom field from AppConfig file.
     */
    public static function addCustomFields($productID, $productFromConfigFile)
    {
        foreach ($productFromConfigFile['config']['custom_fields'] as $customField) {

            /**
             * Represents DB model
             * Relation - id of product
             */
            $customField =
                [
                    'type'         => 'product',
                    'fieldname'    => $customField['name'],
                    'fieldtype'    => $customField['type'],
                    'description'  => $customField['description'],
                    'required'     => $customField['options']['required'],
                    'showorder'    => $customField['options']['show_on_order_form'],
                    'showinvoice'  => $customField['options']['show_on_invoice'],
                    'sortorder'    => $customField['display_order'],
                    'fieldoptions' => $customField['select_options'],
                    'adminOnly'    => $customField['options']['admin_only'],
                    'relid'        => $productID,
                ];

            Capsule::table('tblcustomfields')->insert($customField);
        }
    }

    public static function getProductId($product)
    {
        $product = Capsule::table('tblproducts')->where('name', $product['name'])->first();

        return $product->id;
    }

    public static function getProductFromModTableByName($name)
    {
        $product = Capsule::table('mod_sectigo_products')->where('name', $name)->first();

        return $product;
    }

    public static function createProductGroup($groupName)
    {
        $id = Capsule::table('tblproductconfiggroups')->insertGetId(['name' => $groupName]);
        
        return $id;
    }

    //
    // remaining items
    // entry for configoptions goes into tblpricing
    // how to "enable" the config options on the product
    //
    public static function createProductGroupConfigOption($groupId, $product, $systemCurrencies)
    {
        foreach ($product['config']['configurable_options']['options'] as $option) {
            
            $optionName = $option['name'];

            /**
             * 1 for Dropdown
             * 2 for Radio
             * 3 Yes/No (checkbox)
             * 4 Quantity
             */
            $optionType = self::getTypeOfConfigOptionId($option['type']);

            $configOptionId = Capsule::table('tblproductconfigoptions')->insertGetId(
                [
                    'gid'        => $groupId,
                    'optionname' => $optionName,
                    'optiontype' => $optionType,
                    'qtyminimum' => isset($option['min']) ?  $option['min'] : 0,
                    'qtymaximum' => isset($option['max']) ?  $option['max'] : 0,
                    'order'      => $option['sort_order'],
                    'hidden'     => $option['hidden'] ? 1 : 0,
                ]
            );
            
            // self::createOptionsSubs($configOptionId, $option['values']);

            foreach ($option['values'] as $value) {

                $id = Capsule::table('tblproductconfigoptionssub')->insertGetId(
                    [
                        'configid'   => $configOptionId,
                        'optionname' => $value['name'],
                        'sortorder'  => $value['sort_order'] ? $option['sort_order'] : 0,
                        'hidden'     => 0
                    ]
                );

                if (isset($value['pricing'])) {
                    
                    $base = ['relid' => $id,
                             'type'  => 'configoptions',];
                    
                    $pricing = $value['pricing'];
                    $request = [];

                    foreach ($pricing as $price) {
                        $price = array_merge($base, 
                                            ['currency' => getCurrency($price['currency'], $systemCurrencies)['id'].""],
                                            $price['values']);

                        $request[] = $price;
                    }
                    Capsule::table('tblpricing')->insert($request);
                }
            }      
        }
    }

    public static function linkProductToConfigOptionGroup($groupID, $productID)
    {
        return Capsule::table('tblproductconfiglinks')->insertGetId(['gid' => $groupID, 'pid' => $productID]);
    }

    /**
     * @param array $configOptions 
     */
    public static function deleteSectigoConfigOptions($configOptions)
    {
        foreach ($configOptions as $configOption) {
            Capsule::table('tblproductconfiggroups')->where('name', $configOption)->delete();
        }
    }

    static function insertIncompleteOrder($certType, $moduleType, $orderNumber, $serviceId, $userId)
    {
        Capsule::table('tblsslorders')
            ->insert(['certtype'  => $certType,
                      'module'    => $moduleType,
                      'remoteid'  => $orderNumber,
                      'serviceid' => $serviceId,
                      'status'    => 'Incomplete',
                      'userid'    => $userId,]);
    }

    static function updateSslOrder($sslOrder, $omsNumber)
    {
        Capsule::table('tblsslorders')
            ->where('id', $sslOrder->id)
            ->update(['remoteid'  => $omsNumber]);
    }


    // FIXME
    // check if this function is doing what it needs to do
    //
    public static function getProducts()
    {
        if (!Capsule::schema()->hasTable('mod_sectigo_products')) {
            return [];
        }

        return Capsule::table('mod_sectigo_products')
            ->where('created', true)
            ->get();
    }

    public static function getProductById($id)
    {
        return Capsule::table('tblproducts')->where('id', $id)->first();
    }

    public static function configureProductPricing($productID, $productConfig)
    {
        $currency = Capsule::table('tblcurrencies')->first();;
        $productPricing = [];
        $productPricing['type'] = 'product';
        $productPricing['currency'] = $currency->id;
        $productPricing['relid'] = $productID;
        $productPricing['msetupfee'] = 0;
        $productPricing['qsetupfee'] = 0;
        $productPricing['ssetupfee'] = 0;
        $productPricing['asetupfee'] = 0;
        $productPricing['bsetupfee'] = 0;
        $productPricing['tsetupfee'] = 0;
        $productPricing['monthly'] = -1.00;
        $productPricing['quarterly'] = -1.00;
        $productPricing['semiannually'] = -1.00;
        $productPricing['annually'] = -1.00;
        $productPricing['biennially'] = -1.00;
        $productPricing['triennially'] = -1.00;


        foreach ($productConfig['config']['pricing']['periods'] as $period){
            $productPricing[$period] = 0;
        }

        Capsule::table('tblpricing')->Insert($productPricing);
    }

    private static function getTypeOfConfigOptionId($type)
    {
        /**
         * 1 for Dropdown
         * 2 for Radio
         * 3 Yes/No (checkbox)
         * 4 Quantity
         */
        if($type== 'dropdown')
            return 1;
        if($type== 'radio')
            return 2;
        if($type== 'yes/no')
            return 3;
        if($type== 'checkbox')
            return 3;
        if($type== 'quantity')
            return 4;
    }

    public static function getServicesDetails($email = "", $omsnumber = "", $company = "", $name = "", $surname = "", $sort = 'asc',
                                              $orderBy = 'mod_sectigo_service_details.id', $skip = 0, $take = 15)
    {
        $result = [];
        $query  = Capsule::table('mod_sectigo_service_details')
            ->join('tblhosting', 'mod_sectigo_service_details.serviceid', '=', 'tblhosting.id')
            ->join('tblclients', 'tblhosting.userid', '=', 'tblclients.id')
            ->select(
			'mod_sectigo_service_details.orderid',
			'mod_sectigo_service_details.omsnumber',
			'mod_sectigo_service_details.status',
			'mod_sectigo_service_details.serviceid',
			'mod_sectigo_service_details.created_at',
			'mod_sectigo_service_details.updated_at', 
			'tblclients.firstname', 'tblclients.lastname', 'tblclients.companyname', 'tblclients.email', 'tblhosting.regdate', 'tblclients.id as clientID');
        if ($email) {
            $query->where('tblclients.email', 'like', '%' . $email . '%');
        }
        if ($company) {
            $query->where('tblclients.companyname', 'like', '%' . $company . '%');
        }
        if ($name) {
            $query->where('tblclients.firstname', 'like', '%' . $name . '%');
        }
        if ($surname) {
            $query->where('tblclients.lastname', 'like', '%' . $surname . '%');
        }
        if ($omsnumber) {
            $query->where('mod_sectigo_service_details.omsnumber', 'like', $omsnumber);
        }
        $query->skip($skip)->take($take)->orderBy($orderBy, $sort);

        $rowCount = $query->count();

        $result['data']  = $query->get();
        $result['count'] = $rowCount;

        if ($rowCount < $take) {
            $take = $rowCount;
        }
        $result['currentSet'] = "$skip - $take /  $rowCount";


        if (($skip + $take) < $result['count']) {
            $result['hasNext'] = true;
        } else {
            $result['hasNext'] = false;
        }

        if ($skip == 0) {
            $result['hasPrev'] = false;
        } else {
            $result['hasPrev'] = true;
        }

        return $result;
    }

    public static function deprecateConfigGroup($productId) {
        $name = Capsule::table('tblproductconfiggroups')
            ->join('tblproductconfiglinks', 'tblproductconfiglinks.gid', '=', 'tblproductconfiggroups.id')
            ->where('tblproductconfiglinks.pid', '=', $productId)
            ->value('name');
        $newname = sprintf("%s (pre %s)", $name, (new \DateTime())->format('Y-m-d'));
        Capsule::table('tblproductconfiggroups')
            ->join('tblproductconfiglinks', 'tblproductconfiglinks.gid', '=', 'tblproductconfiggroups.id')
            ->where('tblproductconfiglinks.pid', '=', $productId)
            ->update(['name' => $newname]);
    }

    /**
     * Update pricing on a synchronized product. Product sync is done by using the
     * local API to create a stub product and update values of the existing product
     * with it before deleting the stub. Pricing records are created with the stub
     * product. This method deletes the old product pricing records and points the 
     * newly created ones to the existing product being synced. The pricing table
     * has a generalized relid and a type field to identify what a pricing record
     * is used for. The type values can be one of:
     * 
     *      - product
     *      - addon
     *      - configoptions
     *      - domainaddons
     *      - domainregister
     *      - domaintransfer
     *      - domainrenew
     * 
     * We only really need to be concerned with product and addon. When a sync is
     * performed the product gets a new set of config options and prices are created
     * to go with the new option set, the old ones are left alone for legacy products.
     * 
     * @param int    $from ID of the stub product
     * @param int    $to   ID of the existing product being synced
     * @param string $type the pricing record type to swap.
     */
    public static function swapPricingOptions($from, $to, $type = 'product') {
        // Delete old pricing
        Capsule::table('tblpricing')
            ->where('type', '=', $type)
            ->where('relid', '=', $to)
            ->delete();
        // Switch to new pricing
        Capsule::table('tblpricing')
            ->where('type', '=', $type)
            ->where('relid', '=', $from)
            ->update(['relid' => $to]);
    }

    public static function swapConfigOptions($from, $to) {
        Capsule::table('tblproductconfiglinks')->where('pid', "=", $to)->delete();
        Capsule::table('tblproductconfiglinks')->where('pid', "=", $from)->update(['pid' => $to]);
    }

    public static function swapCustomFields($from, $to) {
        Capsule::table('tblcustomfields')->where('relid', '=', $to)->delete();
        Capsule::table('tblcustomfields')->where('relid', '=', $from)->update(['relid' => $to]);
    }

    public static function deleteProduct($pid) {
        Capsule::table('tblproducts')->where('id', "=", $pid)->delete();
    }

    public static function deleteSectigoModProduct($productName) {
        Capsule::table('mod_sectigo_products')->where('name', '=', $productName)->delete();
    }

    // Get the domain that is used in HTTP & CNAME validation values.
    // This should be pretty static, but it might be worth moving out
    // into some other easier to update place.
    public static function getValidatingDomain() {
        return 'sectigo.com';
    }

    public static function getCustomField($fieldName, $productId) {
        return Capsule::table('tblcustomfields')->where('fieldname', "=", $fieldName)->where('relid', '=', $productId)->first();
    }

    public static function getCustomFieldValue($fieldName, $productId, $serviceId) {
        return Capsule::table('tblcustomfieldsvalues')
                ->join('tblcustomfields', 'tblcustomfieldsvalues.fieldid', '=', 'tblcustomfields.id')
                ->where('tblcustomfields.relid','=', $productId)
                ->where('tblcustomfields.fieldname','=', $fieldName)
                ->where('tblcustomfieldsvalues.relid','=', $serviceId)
                ->value('value');
    }

    // this only works if there is an existing field value set
    public static function setCustomFieldValue($fieldName, $productId, $serviceId, $setVal) {
        Capsule::table('tblcustomfieldsvalues')
            ->join('tblcustomfields', 'tblcustomfieldsvalues.fieldid', '=', 'tblcustomfields.id')
            ->where('tblcustomfields.relid','=', $productId)
            ->where('tblcustomfields.fieldname','=', $fieldName)
            ->where('tblcustomfieldsvalues.relid','=', $serviceId)
            ->update(['tblcustomfieldsvalues.value' => $setVal]);    
    }

    private static function getCustomFieldId($fieldName, $productId) {
        return Capsule::table('tblcustomfields')
            ->where('type', '=', 'product')
            ->where('relid', '=', $productId)
            ->where('fieldname', '=', $fieldName)
            ->value('id');
    }

    public static function createCustomFieldIfDoesNotExist($fieldName, $productId, $adminOnly = '', $required = '') {
        $customFieldId = self::getCustomFieldId($fieldName, $productId);

        if (empty($customFieldId)) self::insertCustomField($fieldName, $productId, $adminOnly, $required);
    }

    public static function upsertCustomFieldValue($fieldName, $productId, $serviceId, $setVal) {
        $id = 0;
        $customFieldId = self::getCustomFieldId($fieldName, $productId);

        $customFieldValues = Capsule::table('tblcustomfieldsvalues')
                                ->where('fieldid', $customFieldId)
                                ->where('relid', $serviceId)
                                ->get();

        $valueCount = count($customFieldValues);

        if ($valueCount == 0) {
            $id = Capsule::table('tblcustomfieldsvalues')->insertGetId(['fieldid' => $customFieldId,'relid' => $serviceId, 'value' => $setVal]);
        } elseif ($valueCount == 1) {
            WHMCSRepository::setCustomFieldValue($fieldName, $productId, $serviceId, $setVal);
            $id = $customFieldValues[0]->id;
        }

        return $id;
    }

    public static function insertCustomFields(array $customFields) {
        Capsule::table('tblcustomfields')->insert($customFields);
    }

    public static function updateProduct($id, $newProductFields) {
        return Capsule::table('tblproducts')->where('id', '=', $id)->update($newProductFields);
    }

    /**
     * This method accepts an integer value for an OMS Number and attempts to map that to a
     * WHMCS service ID. If successful the id is returned. Otherwise it returns 0.
     * 
     * @param int $omsNumber 
     * @return int the service ID associated with the given oms number
     */
    public static function getServiceIdByOMSNumber(int $omsNumber) : int {
        $id = Capsule::table('mod_sectigo_service_details')->where('omsnumber', $omsNumber)->first();
        if(!$id) return 0;
        return $id;
    }

    /**
     * This method wraps the localAPI call to CreateInvoice and on success returns the new invoice ID.
     * 
     * @param array $params the input parameters to CreateInvoice
     * @return int the new invoice ID or 0 if the call fails.
     * @see https://developers.whmcs.com/api-reference/createinvoice/
     */
    public static function createInvoice(array $params) : int {
        $response = localAPI("CreateInvoice", $params);
        $logger = new Logger();
        $logger->error("Create Invoice", $params, $response);
        if ($response['result'] === 'success') {
            return $response["invoiceid"];
        }
        return 0;
    }

    /**
     * Associate the invoice line items of an invoice with a specific service ID. In WHMCS
     * invoice line items are associated with a service by the db fields tblinvoiceitems.relid and
     * tblinvoiceitems.type, however it's not possible to set these view the CreateInvoice local API call. This
     * method takes an invoice id, a service id and a type string (e.g. 'Hosting') and links the invoice items
     * of the invoice to that service.
     * 
     * @param int $invoiceid the invoice ID to update
     * @param int $serviceid the ID of the service to associate with the invoice
     * @param string $type the value to assign to tblinvoiceitems.type
     * @return bool true if update was successful, otherwise false
     */
    public static function attachInvoiceToService(int $invoiceid, int $serviceid, string $type) : bool {
        // First look up the invoice and make sure it exists
        $logger = new Logger();
        $logger->error("Attaching item to invoice", $invoiceid, $serviceid);
        $invoice = Capsule::table("tblinvoices")->where("id", $invoiceid)->first();
        if(!$invoice) return false;
        $logger->error("Getting service details", $invoiceid, $serviceid);
        // Then look up the service and make sure it exists
        $service = self::getServiceDetail($serviceid);
        $logger->error("service details", $serviceid, $service);
        if(!$service) return false;
        // Make sure the invoice and service belong to the same user
        $logger->error("Userid check", $invoice->userid, $service->userid);
        if(!$invoice->userid == $service->userid) return false;
        Capsule::table("tblinvoiceitems")
            ->where('invoiceid', $invoiceid)
            ->update([
                'relid' => $serviceid,
                'type'  => $type
            ]);
        return true;
    }

    /**
     * Return an array of invoice items for extendable ssl products
     * 
     * @param int $invoiceid The ID of the invoice to scan
     * @return array of objects from tblinvoiceitems
     */
    public static function getExtendableItems(int $invoiceid) {
        $rv = Capsule::table("tblinvoiceitems")
                    ->join('tblhosting', 'tblinvoiceitems.relid', '=', 'tblhosting.id')
                    ->join('tblproducts', 'tblhosting.packageid', '=', 'tblproducts.id')
                    ->where('tblinvoiceitems.type', 'HostingExtend')
                    ->where('tblproducts.paytype', 'onetime')
                    ->where('tblinvoiceitems.invoiceid', $invoiceid)
                    ->get();
        return $rv;
    }

    /**
     * Return the tax status of the product
     * 
     * @param int $serviceid the service id to check
     * @return bool true if the product is taxable
     */
    public static function isServiceTaxable(int $serviceid) : bool {
        $result = Capsule::table("tblproducts")
                        ->join("tblhosting", "tblproducts.id", '=', "tblhosting.packageid")
                        ->where("tblhosting.id", $serviceid)->first();
        if($result->tax) return true;
        return false;
    }

    public static function insertCustomField($fieldName, $productId, $adminOnly = '', $required = '') {
        $customFieldId = self::getCustomFieldId($fieldName, $productId);

        if (!$customFieldId) {
            $fieldValues = [
                'type' => 'product',
                'relid' => $productId,
                'fieldname' => $fieldName,
                'fieldtype' => 'text',
                'description' => '',
                'fieldoptions' => '',
                'regexpr' => '',
                'adminonly' => $adminOnly,
                'required' => $required,
                'showorder' => '',
                'showinvoice' => '',
            ];
            Capsule::table('tblcustomfields')
                ->insert($fieldValues);

            $customFieldId = self::getCustomFieldId($fieldName, $productId);
        }

        return $customFieldId;
    }

    public static function saveFormFieldValues($productId, $serviceId, $values) {
        if ($serviceId && $productId) {
            $customField = self::getCustomField(self::FORM_FIELD_VALUES_NAME, $productId);
            if (!$customField) {
                self::insertCustomField(self::FORM_FIELD_VALUES_NAME, $productId, 'on');
            }
            self::upsertCustomFieldValue(self::FORM_FIELD_VALUES_NAME, $productId, $serviceId, Helpers::encrypt($values));
        }
    }

    public static function loadFormFieldValues($productId, $serviceId) {

        $formFieldValues = $serviceId && $productId
            ? json_decode(
                Helpers::decrypt(
                    self::getCustomFieldValue(self::FORM_FIELD_VALUES_NAME, $productId, $serviceId)
                ),
                true)
            : [];

        if (!empty($formFieldValues)) {
            foreach ($formFieldValues as $field => $value) {
                if (strtolower(gettype($value)) == 'array') {
                    $formFieldValues[$field] = implode(',', $value);
                }
            }
        } else {
            $formFieldValues = [];
        }

        return $formFieldValues;
    }
}
